# Module of TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2008 Sven Dowideit, SvenDowideit@home.org.au
# and TWiki Contributors. All Rights Reserved. TWiki Contributors
# are listed in the AUTHORS file in the root of this distribution.
# NOTE: Please extend that file, not this notice.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.

=pod

---+ package TWiki::LoginManager::TracOnTWikiLogin

This is a login manager that you can specify in the security setup section of [[%SCRIPTURL%/configure%SCRIPTSUFFIX%][configure]]. It provides users with a template-based form to enter usernames and passwords, and works with the PasswordManager that you specify to verify those passwords.

Subclass of TWiki::LoginManager; see that class for documentation of the
methods of this class.

=cut

package TWiki::LoginManager::TracOnTWikiLogin;

use strict;
use Assert;
use DBI;
use TWiki::LoginManager::TemplateLogin;

@TWiki::LoginManager::TracOnTWikiLogin::ISA = ('TWiki::LoginManager::TemplateLogin');

sub new {
    my ( $class, $session ) = @_;

    my $this = bless( $class->SUPER::new($session), $class );
    $session->enterContext('can_login');
    return $this;
}

sub finish {
    my $this = shift;
    
    $this->{DB}->disconnect() if defined($this->{DB});
    undef $this->{DB};
    
    $this->SUPER::finish();
}

=pod

---++ ObjectMethod userLoggedIn( $login, $wikiname)

Only updates the trac database on login - so if a user changes their email address
they will need to logout and log back in again

=cut

sub userLoggedIn {
    my ( $this,  $login, $wikiname)  = @_;
    my $twiki = $this->{twiki};
    my $query = $twiki->{cgiQuery};

    ASSERT( $this->isa('TWiki::LoginManager::TracOnTWikiLogin') ) if DEBUG;
    $this->SUPER::userLoggedIn( $login, $wikiname);

    my $tracSessionCookieName = 'trac_session';
    my $tracAuthCookieName = 'trac_auth';

    #use guest user so that login can be used to force a db refresh    
    $login = $TWiki::cfg{DefaultUserLogin} unless (defined($login));
    if ($login eq $TWiki::cfg{DefaultUserLogin}) {
        #expire the cookies
        $this->addTracCookie($tracSessionCookieName, '');
        $this->addTracCookie($tracAuthCookieName, '');
        return;
    }
    
    #note there is also a trac_form_token
    # see if there is a trac UID cookie
    my %cookies = fetch CGI::Cookie;
    if (defined( $cookies{$tracAuthCookieName} ) ) {
        #TODO: check that the cookie is for _this_ logged in user - if not, change the cookie
#        my $cookieValue = ....; - need to look up db (might be better for performance not to.)
#        unless ($cookies{$tracAuthCookieName}->value eq $cookieValue) {
#            $this->addTracCookie($tracAuthCookieName, $cookieValue);
#            $this->updateTracDB($login, $cookieValue);
#        }
    } else {
        #add a cookie for this user
        my @chars = ( "A" .. "Z", "a" .. "z", 0 .. 9 );
        my $cookieValue = join("", @chars[ map { rand @chars } ( 1 .. 32 ) ]);
        $this->addTracCookie($tracAuthCookieName, $cookieValue);
        $this->updateTracDB($login, $cookieValue);
    }
}

=pod

---++ ObjectMethod updateTracDB( $tracCookieName, $login )



=cut

sub updateTracDB {
    my ($this, $login, $cookieValue) = @_;
    my $twiki = $this->{twiki};
    my $query = $twiki->{cgiQuery};
    
    return unless (defined($twiki->{users}));
        
    my $cUID = $twiki->{users}->getCanonicalUserID($login);
    my @emails = $twiki->{users}->getEmails($cUID);
    my $name = $twiki->{users}->getWikiName($cUID); #TODO: i wonder what happens if this is a html href?
    my $time = time();
        
    #if their name & email is not in the Trac DB, add it.
    my $sqldb = $this->getDB();
    my $st = $sqldb->prepare("SELECT * from session_attribute WHERE sid = ?");
    $st->bind_param(1, $login);
    $st->execute( );
    my $hash_ary = $st->fetchall_arrayref();
    my $sessionAttributeStatement;
    my $sessionStatement;
    if (defined($$hash_ary[0])) {
        #there's a user in the db already
        $sessionAttributeStatement = $sqldb->prepare("UPDATE session_attribute Set authenticated = ?, value = ? WHERE sid = ? AND name = ?");
        $sessionStatement = $sqldb->prepare("UPDATE session Set authenticated = ?, last_visit = ? WHERE sid = ? ");
    } else {
        $sessionAttributeStatement = $sqldb->prepare("INSERT INTO session_attribute (authenticated, value, sid, name) VALUES(?, ?, ?, ?);");
        $sessionStatement = $sqldb->prepare("INSERT INTO session (authenticated, last_visit, sid) VALUES(?, ?, ?);");
    }
    $sessionAttributeStatement->bind_param(1, 1);
    $sessionAttributeStatement->bind_param(2, $name);
    $sessionAttributeStatement->bind_param(3, $login);
    $sessionAttributeStatement->bind_param(4, 'name');
    $sessionAttributeStatement->execute( );     
    $sessionAttributeStatement->bind_param(1, 1);
    $sessionAttributeStatement->bind_param(2, $emails[0] || '');
    $sessionAttributeStatement->bind_param(3, $login);
    $sessionAttributeStatement->bind_param(4, 'email');
    $sessionAttributeStatement->execute( );  
    $sessionStatement->bind_param(1, 1);
    $sessionStatement->bind_param(2, $time);
    $sessionStatement->bind_param(3, $login);
    $sessionStatement->execute( );
    
    #TODO: work out whats supposed to happen for 'authenticated' columns
    my $ipnr = $query->remote_host();
    my $st3 = $sqldb->prepare("SELECT * from auth_cookie WHERE name = ? AND ipnr = ?");
    $st3->bind_param(1, $login);
    $st3->bind_param(2, $ipnr);
    $st3->execute( );
    $hash_ary = $st3->fetchall_arrayref();

    #set the authenticated cookie for this session
    my $authCookieStatement;
    if (defined($$hash_ary[0])) {
        $authCookieStatement = $sqldb->prepare("UPDATE auth_cookie Set cookie = ? , time= ? WHERE name = ? AND ipnr = ?;");
    } else {
        $authCookieStatement = $sqldb->prepare("INSERT INTO auth_cookie (cookie, time, name, ipnr) VALUES(?, ?, ?, ?);");
    }
    $authCookieStatement->bind_param(1, $cookieValue);
    $authCookieStatement->bind_param(2, $time);
    $authCookieStatement->bind_param(3, $login);
    $authCookieStatement->bind_param(4, $ipnr);
    $authCookieStatement->execute( );
    
    #add or remove the user from the TWikiAdminGroup
    my $inPermissionStatement = $sqldb->prepare("SELECT * from permission WHERE username = ? AND action = ?");
    $inPermissionStatement->bind_param(1, $login);
    $inPermissionStatement->bind_param(2, 'TRAC_ADMIN');
    $inPermissionStatement->execute( );
    $hash_ary = $inPermissionStatement->fetchall_arrayref();

    #set the authenticated cookie for this session
    my $setPermissionStatement;
    if (defined($$hash_ary[0]) && !$twiki->{users}->isAdmin($cUID)) {
        #if the user has lost admin rights
        $setPermissionStatement = $sqldb->prepare("DELETE FROM permission WHERE username = ? AND action = ?;");
    } 
    if (!defined($$hash_ary[0]) && $twiki->{users}->isAdmin($cUID)) {
        #if the user has gained admin rights
        $setPermissionStatement = $sqldb->prepare("INSERT INTO permission (username, action) VALUES(?, ?);");
    }
    if (defined($setPermissionStatement)) {
        $setPermissionStatement->bind_param(1, $login);
        $setPermissionStatement->bind_param(2, 'TRAC_ADMIN');
        $setPermissionStatement->execute( );
    }    
    #TODO: or should we populate _all_ groups..
    #eachMembership($user) -> $iterator
}


=pod

---++ ObjectMethod addTracCookie( $tracCookieName, $value )

if $value is set to an empty string, or undef, then the cookie is expired.

=cut

sub addTracCookie {
    my ($this, $tracCookieName, $value) = @_;
    
    my $domain = $TWiki::cfg{Plugins}{TracOnTWikiContrib}{Domain};
    $domain = $TWiki::cfg{DefaultUrlHost} unless (defined($domain) && ($domain ne ''));
    #TODO: crap - the URI stuff kills us here
    #$domain = $twiki->{urlHost} unless (defined($domain) && ($domain ne ''));
    #print STDERR "wanting to add cookie $tracCookieName == ($value)for host $domain";


    #really really should use URI
    $domain =~ m!^([^:]*://)([^/]*)(/.*)?$!;
    $domain = $2;
    my $path = $3;
    
    my $expires = '+7d';
    if (!defined($value) || $value eq '') {
        $expires = '-1000d';    #expire cookie
    }
    
    my $tracCookie = new CGI::Cookie(
                -name=>$tracCookieName,
                -value=>$value,
#TODO: restrict to allowed domain
#                        -domain=>'10.10.10.242',
                        -domain=>$domain,
                        -path=>$path,
                        -expires=>$expires
            );
    $this->addCookie($tracCookie);
    #print STDERR "adding cookie $tracCookieName for host $domain";
    
}

###############################################################################
#DB access methods


#todo: cache DB connections
sub getDB {
    my ( $this ) = @_;
#    ASSERT($this->isa( 'TWiki::Users::JoomlaUserMapping')) if DEBUG;
    my ( $dbi_dsn, $dbi_user, $dbi_passwd ) = 
            ($TWiki::cfg{Plugins}{TracOnTWikiContrib}{DBI_dsn}, 
            $TWiki::cfg{Plugins}{TracOnTWikiContrib}{DBI_username}, 
            $TWiki::cfg{Plugins}{TracOnTWikiContrib}{DBI_password});

        	
    unless (defined($this->{DB})) {
        $this->{DB} = DBI->connect( $dbi_dsn, $dbi_user, $dbi_passwd );
    }
    return $this->{DB};
}

1;
